@@ -1,7 +1,6 @@
 import dis
 import inspect
 import opcode as _opcode
-import stat
 import struct
 import sys
 import types
@@ -26,6 +25,11 @@ import bytecode as _bytecode
 from bytecode.flags import CompilerFlags
 from bytecode.instr import (
     _UNSET,
+    BITFLAG2_INSTRUCTIONS,
+    BITFLAG_INSTRUCTIONS,
+    INTRINSIC,
+    INTRINSIC_1OP,
+    INTRINSIC_2OP,
     PLACEHOLDER_LABEL,
     UNSET,
     BaseInstr,
@@ -35,12 +39,15 @@ from bytecode.instr import (
     Instr,
     InstrArg,
     InstrLocation,
+    Intrinsic1Op,
+    Intrinsic2Op,
     Label,
     SetLineno,
     TryBegin,
     TryEnd,
     _check_arg_int,
     const_key,
+    opcode_has_argument,
 )
 
 # - jumps use instruction
@@ -90,7 +97,7 @@ class ConcreteInstr(BaseInstr[int]):
         super().__init__(name, arg, lineno=lineno, location=location)
 
     def _check_arg(self, name: str, opcode: int, arg: int) -> None:
-        if opcode >= _opcode.HAVE_ARGUMENT:
+        if opcode_has_argument(opcode):
             if arg is UNSET:
                 raise ValueError("operation %s requires an argument" % name)
 
@@ -126,11 +133,17 @@ class ConcreteInstr(BaseInstr[int]):
         return (self._location, self._name, self._arg)
 
     def get_jump_target(self, instr_offset: int) -> Optional[int]:
+        # When a jump arg is zero the jump always points to the first non-CACHE
+        # opcode following the jump. The passed in offset is the offset at
+        # which the jump opcode starts. So to compute the target, we add to it
+        # the instruction size (accounting for extended args) and the
+        # number of caches expected to follow the jump instruction.
+        s = (
+            (self._size // 2) if OFFSET_AS_INSTRUCTION else self._size
+        ) + self.use_cache_opcodes()
         if self.is_forward_rel_jump():
-            s = (self._size // 2) if OFFSET_AS_INSTRUCTION else self._size
             return instr_offset + s + self._arg
         if self.is_backward_rel_jump():
-            s = (self._size // 2) if OFFSET_AS_INSTRUCTION else self._size
             return instr_offset + s - self._arg
         if self.is_abs_jump():
             return self._arg
@@ -156,13 +169,21 @@ class ConcreteInstr(BaseInstr[int]):
     def disassemble(cls: Type[T], lineno: Optional[int], code: bytes, offset: int) -> T:
         index = 2 * offset if OFFSET_AS_INSTRUCTION else offset
         op = code[index]
-        if op >= _opcode.HAVE_ARGUMENT:
+        if opcode_has_argument(op):
             arg = code[index + 1]
         else:
             arg = UNSET
         name = _opcode.opname[op]
         return cls(name, arg, lineno=lineno)
 
+    def use_cache_opcodes(self) -> int:
+        return (
+            # Not supposed to be used but we need it
+            dis._inline_cache_entries[self._opcode]  # type: ignore
+            if sys.version_info >= (3, 11)
+            else 0
+        )
+
 
 class ExceptionTableEntry:
     """Entry for a given line in the exception table.
@@ -220,7 +241,6 @@ class ExceptionTableEntry:
 
 
 class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLineno]]):
-
     #: List of "constant" objects for the bytecode
     consts: List
 
@@ -297,7 +317,6 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
     def from_code(
         code: types.CodeType, *, extended_arg: bool = False
     ) -> "ConcreteBytecode":
-
         instructions: MutableSequence[Union[SetLineno, ConcreteInstr]]
         # For Python 3.11+ we use dis to extract the detailed location information at
         # reduced maintenance cost.
@@ -308,7 +327,9 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
                 ConcreteInstr(
                     i.opname,
                     i.arg % 256 if i.arg is not None else UNSET,
-                    location=InstrLocation.from_positions(i.positions),
+                    location=InstrLocation.from_positions(i.positions)
+                    if i.positions
+                    else None,
                 )
                 for i in dis.get_instructions(code, show_caches=True)
             ]
@@ -470,7 +491,6 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
         # then use as many 254 offset with no line change to reduce the offset to
         # less than 254.
         if doff > 254:
-
             linetable.append(struct.pack("Bb", 254, dlineno))
             doff -= 254
 
@@ -692,7 +712,6 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
 
         # We track the last set lineno to be able to compute deltas
         for _, i_size, new_lineno, location in iter_in:
-
             # Infer the line if location is None
             location = location or InstrLocation(new_lineno, None, None, None)
 
@@ -743,7 +762,7 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
                 continue
 
             if extended_arg is not None:
-                arg = (extended_arg << 8) + instr.arg
+                arg = UNSET if instr.name == "NOP" else (extended_arg << 8) + instr.arg
                 extended_arg = None
 
                 instr = ConcreteInstr(
@@ -799,9 +818,10 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
         while value:
             temp.append(value & 63 | (64 if temp else 0))
             value >>= 6
+        temp = temp or [0]
         if set_begin_marker:
             temp[-1] |= 128
-        return reversed(temp or [0])
+        return reversed(temp)
 
     def _assemble_exception_table(self) -> bytes:
         table = bytearray()
@@ -944,7 +964,7 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
         # array.
         # As a consequence, the instruction argument can be either:
         # - < len(varnames): the name is shared an we can directly use
-        #   the index to access the name in celvars
+        #   the index to access the name in cellvars
         # - > len(varnames): the name is not shared and is offset by the
         #   number unshared varname.
         # Free vars are never shared and correspond to index larger than the
@@ -1007,8 +1027,10 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
                 elif c_instr.opcode in _opcode.haslocal:
                     arg = self.varnames[c_arg]
                 elif c_instr.opcode in _opcode.hasname:
-                    if sys.version_info >= (3, 11) and c_instr.name == "LOAD_GLOBAL":
+                    if c_instr.name in BITFLAG_INSTRUCTIONS:
                         arg = (bool(c_arg & 1), self.names[c_arg >> 1])
+                    elif c_instr.name in BITFLAG2_INSTRUCTIONS:
+                        arg = (bool(c_arg & 1), bool(c_arg & 2), self.names[c_arg >> 2])
                     else:
                         arg = self.names[c_arg]
                 elif c_instr.opcode in _opcode.hasfree:
@@ -1019,7 +1041,13 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
                         name = self.freevars[c_arg - ncells]
                         arg = FreeVar(name)
                 elif c_instr.opcode in _opcode.hascompare:
-                    arg = Compare(c_arg)
+                    arg = Compare(
+                        (c_arg >> 4) if sys.version_info >= (3, 12) else c_arg
+                    )
+                elif c_instr.opcode in INTRINSIC_1OP:
+                    arg = Intrinsic1Op(c_arg)
+                elif c_instr.opcode in INTRINSIC_2OP:
+                    arg = Intrinsic2Op(c_arg)
                 else:
                     arg = c_arg
 
@@ -1066,7 +1094,6 @@ class ConcreteBytecode(_bytecode._BaseBytecodeList[Union[ConcreteInstr, SetLinen
 
 
 class _ConvertBytecodeToConcrete:
-
     # XXX document attributes
 
     #: Default number of passes of compute_jumps() before giving up.  Refer to
@@ -1117,7 +1144,6 @@ class _ConvertBytecodeToConcrete:
         free_instrs: List[int] = []
 
         for instr in self.bytecode:
-
             # Enforce proper use of CACHE opcode on Python 3.11+ by checking we get the
             # number we expect or directly generate the needed ones.
             if isinstance(instr, Instr) and instr.name == "CACHE":
@@ -1196,7 +1222,7 @@ class _ConvertBytecodeToConcrete:
                 assert isinstance(arg, str)
                 arg = self.add(self.varnames, arg)
             elif instr.opcode in _opcode.hasname:
-                if sys.version_info >= (3, 11) and instr.name == "LOAD_GLOBAL":
+                if instr.name in BITFLAG_INSTRUCTIONS:
                     assert (
                         isinstance(arg, tuple)
                         and len(arg) == 2
@@ -1205,8 +1231,18 @@ class _ConvertBytecodeToConcrete:
                     ), arg
                     index = self.add(self.names, arg[1])
                     arg = int(arg[0]) + (index << 1)
+                elif instr.name in BITFLAG2_INSTRUCTIONS:
+                    assert (
+                        isinstance(arg, tuple)
+                        and len(arg) == 3
+                        and isinstance(arg[0], bool)
+                        and isinstance(arg[1], bool)
+                        and isinstance(arg[2], str)
+                    ), arg
+                    index = self.add(self.names, arg[2])
+                    arg = int(arg[0]) + 2 * int(arg[1]) + (index << 2)
                 else:
-                    assert isinstance(arg, str)
+                    assert isinstance(arg, str), f"Got {arg}, expected a str"
                     arg = self.add(self.names, arg)
             elif instr.opcode in _opcode.hasfree:
                 if isinstance(arg, CellVar):
@@ -1218,6 +1254,14 @@ class _ConvertBytecodeToConcrete:
                     arg = self.bytecode.freevars.index(arg.name)
             elif instr.opcode in _opcode.hascompare:
                 if isinstance(arg, Compare):
+                    # In Python 3.12 the 4 lowest bits are used for caching
+                    # See compare_masks in compile.c
+                    if sys.version_info >= (3, 12):
+                        arg = arg._get_mask() + (arg.value << 4)
+                    else:
+                        arg = arg.value
+            elif instr.opcode in INTRINSIC:
+                if isinstance(arg, (Intrinsic1Op, Intrinsic2Op)):
                     arg = arg.value
 
             # The above should have performed all the necessary conversion
@@ -1228,7 +1272,7 @@ class _ConvertBytecodeToConcrete:
 
             # If the instruction expect some cache
             if sys.version_info >= (3, 11):
-                self.required_caches = dis._inline_cache_entries[c_instr.opcode]
+                self.required_caches = c_instr.use_cache_opcodes()
                 self.seen_manual_cache = False
 
             self.instructions.append(c_instr)
@@ -1289,6 +1333,12 @@ class _ConvertBytecodeToConcrete:
             target_index = self.labels[label]
             target_offset = label_offsets[target_index]
 
+            # FIXME use opcode
+            # Under 3.12+, FOR_ITER, SEND jump is increased by 1 implicitely
+            # to skip over END_FOR, END_SEND see Python/instrumentation.c
+            if sys.version_info >= (3, 12) and instr.name in ("FOR_ITER", "SEND"):
+                target_offset -= 1
+
             if instr.is_forward_rel_jump():
                 instr_offset = label_offsets[index]
                 target_offset -= instr_offset + (
@@ -1315,7 +1365,6 @@ class _ConvertBytecodeToConcrete:
 
         # Resolve labels for exception handling entries
         for tb, entry in self.exception_handling_blocks.items():
-
             # Set the offset for the start and end offset from the instruction
             # index stored when assembling the concrete instructions.
             entry.start_offset = instruction_offsets[entry.start_offset]
